@@ -6,6 +6,12 @@ E_keV = 15.2
 q_range = (10,35)
 # path to crystal cif file
 cifPath = 'analysis/BaCO3.cif'
+
+odf_mode = 'hsh' # 'grid' # 
+grid_resolution = 15 # degree # ignored if hsh
+hsh_max_order = 12 # ignored if gridbased
+
+# classic mode parameters
 # approximate crystal size along (a,b,c)-axes in nm
 crystalsize = (15,15,15)
 # angular sampling
