@@ -4,11 +4,11 @@ transverse_horizontal = (0,1,0) # j in mumott
 transverse_vertical = (0,0,1) # k in mumott
 ################################ don't change
 
-# detector geometry:
-flip_detector_ud=False
+# detector geometry: 
+flip_detector_ud=False 
 flip_detector_lr=False
-detector_direction_origin = (0,-1,0)
-detector_direction_positive_90 = (0,0,-1)
+detector_direction_origin = (0,-1,0) # this conforms to standard pyFAI output
+detector_direction_positive_90 = (0,0,-1) # this conforms to standard pyFAI output
 
 # sample movements:
 inner_axis = (0,0,1) # inner rotation axis
