@@ -2,31 +2,36 @@ import os, glob, sys, shutil
 os.environ["MKL_NUM_THREADS"] = "1" 
 os.environ["NUMEXPR_NUM_THREADS"] = "1" 
 os.environ["OMP_NUM_THREADS"] = "1"
+os.environ["OPENBLAS_NUM_THREADS"] = "1"
 import numpy as np
 import h5py, hdf5plugin
 import matplotlib.pyplot as plt
+from matplotlib import colors
 from importlib import reload # this is to reload
 import imageio.v3 as iio
 from time import time
 from typing import Union, Optional
 
-from .src import pyfai_plugins as pyf
-from .src import mumott_plugins as mum
-from .src.model import model
-from .src import data as dat
-from .src.fitting import fitting
+from .src.integration import pyfai_plugins as pyf
+from .src.data_treatment import mumott_plugins as mum
+from .src.model.model_textom import model_textom
+from .src.data_treatment import data as dat
+from .src.data_treatment import baselines as bln
+from .src.optimization.fitting import fitting
 from .src import handle as hdl
 from .src import plotting as plot
-from .src import optimizer 
-from .src import symmetries as sym
-from .src import mask as msk
+from .src.optimization import optimizer 
+from .src.model import symmetries as sym
+from .src.data_treatment import mask as msk
 from .src import misc as msc
-from .src import orix_plugins as orx
+from .src.analysis import orix_plugins as orx
 from .src import exceptions as exc
-from .src import rotation as rot
-from .src import segmentation as seg
-from .src import orderparameters as ord
-from .src import absorption_corrector as abc
+from .src.model import rotation as rot
+from .src.analysis import segmentation as seg
+from .src.analysis import orderparameters as ord
+from .src.data_treatment import absorption_corrector as abc
+from .src.model import model_crystal as cry
+from .src.creator import creator
 from .config import use_gpu, n_threads, data_type
 
 sample_dir = os.getcwd()
@@ -72,7 +77,48 @@ def check_state( ):
         except:
             print(f'\tNo {co[0]} found')
     
-def integrate( parallel=True, wait=5., confirm=True, ignore=[] ):
+def integration_test(dset_no=0):
+    """Set up integration parameters, perform a 2D and 1D integration and plot them
+
+    Parameters
+    ----------
+    dset_no : int, optional
+        allows to choose a different file from the raw data folder, by default 0
+    """
+    test_filenames = pyf.start_test_integration(sample_dir, dset_no=dset_no)
+    print('\tWrote test data: '+test_filenames)
+    f,ax = plt.subplots(2, sharex=True, figsize=(6,8))
+    with h5py.File(test_filenames[0], 'r') as hf:
+        q = hf['radial_units'][()]
+        chi = hf['azimuthal_units'][()]
+        data = hf['cake_integ'][()]
+    ax[0].pcolormesh(*np.meshgrid(q,chi), data.max(axis=0), norm=colors.LogNorm())
+    ax[0].set_ylabel('chi / degree')
+    with h5py.File(test_filenames[1], 'r') as hf:
+        q = hf['radial_units'][()]
+        data = hf['cake_integ'][()]
+    ax[1].plot(q.flatten(), data.max(axis=0))
+    ax[1].set_ylabel('Intensity')
+    ax[1].set_xlabel('q / nm^-1')
+    f.tight_layout()
+
+def generate_sample():
+
+    global mod
+    gen_path_sample = os.path.join(sample_dir,'analysis','generation.py')
+    if not os.path.isfile( gen_path_sample ):
+        gen_path_module = hdl.get_file_path('textom',os.path.join('input','generation.py'))
+        os.makedirs(os.path.join(sample_dir,'analysis'), exist_ok=True )
+        shutil.copyfile(gen_path_module, gen_path_sample )
+    hdl.open_with_editor(gen_path_sample)
+    
+    crt = creator(sample_dir)
+    crt.setup_generated_sample()
+    crt.define_coefficients()
+    crt.save_projections()
+    
+
+def integrate( parallel=True, wait=5., confirm=True, ignore=[], parallel_gpu=False ):
     """Integrates raw 2D diffraction data via pyFAI
 
     All necessary input will be handled via the file integration_parameters.py
@@ -92,21 +138,22 @@ def integrate( parallel=True, wait=5., confirm=True, ignore=[] ):
         here you can provide a list of datasets you want to skip (if they crash
         they will automatically be skipped)
     """
-    # if mode=='parallel':
-    #     pyf.start_integration_parallel( sample_dir, confirm=confirm, ignore=ignore )
-    # elif mode=='online':
-    pyf.start_integration_online( sample_dir, wait=wait, confirm=confirm, ignore=ignore, parallel=parallel )
-    # else:
-    #     print('Integration mode not recognized')
+    if parallel:
+        pyf.start_integration_parallel( sample_dir, confirm=confirm, ignore=ignore )
+    else:
+        pyf.start_integration_online( sample_dir, wait=wait, confirm=confirm, ignore=ignore, parallel=parallel_gpu )
 
-def check_geometry(testfile_path,testfile_h5path,frame_no,
+def check_geometry(testfile_path,testfile_h5path,dset_no=0,
                     vmin=1, vmax = 10, logscale = False
-                    ):
-    
+                    ):     
+    #####
+    # i would like to make the first 2 arguments auto, maybe also this function could replace "integrate_test"
+    ####
+
     pyf.check_detector_geometry(
         os.path.join(sample_dir,'integration_parameters.py'),
         testfile_path,testfile_h5path,
-        frame_no,
+        dset_no,
         os.path.join(sample_dir,'analysis','geometry.py'),
         vmin,vmax,logscale,
         )
@@ -162,9 +209,9 @@ def align_data(
     geo_path_sample = os.path.join(sample_dir,'analysis','geometry.py')
     if not os.path.isfile( geo_path_sample ):
         geo_path_module = hdl.get_file_path('textom',os.path.join('input','geometry.py'))
-        hdl.open_with_editor(geo_path_module)
         os.makedirs(os.path.join(sample_dir,'analysis'), exist_ok=True )
         shutil.copyfile( geo_path_module, geo_path_sample )
+    hdl.open_with_editor(geo_path_sample)
     # do data preprocessing if not yet done
     if ( os.path.isfile(os.path.join(sample_dir,'analysis/data_mumott.h5')) and
         not redo_import ):
@@ -244,12 +291,14 @@ def check_alignment_projection( g=0 ):
 #         for g in G:
 #             fid.write(f'{g}\n')
 
-def reconstruct_1d_full( redo_import=False, only_mumottize=False, batch_size=10 ):
+def reconstruct_1d_full( q_index_range=None, redo_import=False, only_mumottize=False, batch_size=10 ):
     """Reconstructs scalar tomographic data such as azimutally averaged
     diffraction data. Uses the same alignment as textom
 
     Parameters
-    ----------
+    ----------    
+    q_index_range : list or ndarray or None, optional
+        [starting_q_index, end_q_index], takes the whole range if None, by default None
     redo_import : bool, optional
         _description_, by default False
     only_mumottize : bool, optional
@@ -263,16 +312,16 @@ def reconstruct_1d_full( redo_import=False, only_mumottize=False, batch_size=10
     input_path_sample = os.path.join(sample_dir,'analysis','rec_1d','input_1Drec.py')
     if not os.path.isfile( input_path_sample ):
         input_path_module = hdl.get_file_path('textom',os.path.join('input','input_1Drec.py'))
-        hdl.open_with_editor(input_path_module) # take the one from the textom module
         shutil.copyfile(input_path_module, input_path_sample ) # copy to the sample directory
     else:
         msc.cp_add_dt(input_path_sample, sample_dir, now=False) # save the old version with modification date
-        hdl.open_with_editor(input_path_sample) # edit and use the same file
+    hdl.open_with_editor(input_path_sample) # edit and use the same file
 
     datapath = os.path.join(sample_dir,'analysis','rec_1d','data_rec1d.h5')
     if not os.path.isfile(datapath) or redo_import:
         datapath=None
     out_path = mum.backproject_all( sample_dir, input_path_sample, datapath, 
+                            range_q=q_index_range,
                             only_mumottize=only_mumottize, batch_size=batch_size)
     with h5py.File(os.path.join(sample_dir,'analysis','projectors.h5'), 'r') as hf:
         mask = hf['mask_voxels'][()]
@@ -292,40 +341,16 @@ def make_model( light_mode=False ):
     crystal_path_sample = os.path.join(sample_dir,'analysis','crystal.py')
     if not os.path.isfile( crystal_path_sample ):
         crystal_path_module = hdl.get_file_path('textom',os.path.join('input','crystal.py'))
-        hdl.open_with_editor(crystal_path_module)
         os.makedirs(os.path.join(sample_dir,'analysis'), exist_ok=True )
         shutil.copyfile(crystal_path_module, crystal_path_sample )
-    mod = model( sample_dir, light=light_mode )
-    #mod.get_orders(n_max=12)
+        hdl.open_with_editor(crystal_path_sample)
 
-def check_diffractlets(order=4, lam=0):
-    """plots 2D diffractlets for checking angular sampling
-
-    Parameters
-    ----------
-    order : int, optional
-        principal HSH order, by default 4
-    lam : int, optional
-        running number within the sHSHs, choose whatever, by default 0
-
-    Returns
-    -------
-    _type_
-        _description_
-    """
-    global mod
-    mod.set_orders(order)
-    l = np.clip( mod.Sn[order][0] + lam,
-        mod.Sn[order][0],mod.Sn[order][1]-1)
-    plt.figure()
-    plt.matshow(mod.difflets[l].reshape(mod.detShape))
-    plt.xlabel('chi')
-    plt.ylabel('q')
+    mod = model_textom( sample_dir, light=light_mode )
 
 def list_sample_rotations():
     print('Index\tInner\tOuter')
     for g in range(mod.Omega.size):
-        print(f'{g}\t{np.round(mod.Omega[g]*180/np.pi,1)}\t{np.round(mod.Kappa[g]*180/np.pi,1)}')
+        print(f'{g}\t{np.round(mod.Omega[g],1)}\t{np.round(mod.Kappa[g],1)}')
 
 def show_sample_outline( ):
     with h5py.File(os.path.join(sample_dir,'analysis','projectors.h5'), 'r') as hf:
@@ -333,8 +358,8 @@ def show_sample_outline( ):
         mask_voxels = hf['mask_voxels'][()]
     msk.check_tomogram(tomogram, mask_voxels)
 
-def check_peak_regions( projection=0, datafile_pattern='.h5' ):
-    """shows the selected peak regions and with the average of a selected projection, maybe give a redo-option
+def mask_peak_regions( projection=0, datafile_pattern='.h5' ):
+    """lets the user choose the regions that contain the peaks for fitting
 
     Parameters
     ----------
@@ -345,24 +370,92 @@ def check_peak_regions( projection=0, datafile_pattern='.h5' ):
     """
     filelist = dat.get_data_list(sample_dir, datafile_pattern)
     with h5py.File(os.path.join(sample_dir,'data_integrated',filelist[projection]), 'r') as hf:
-        powder = hf['cake_integ'][()].sum(axis=0)
-        q_powder = np.flatten(hf['radial_units'][()])
-    qmask = os.path.join(sample_dir,'analysis','peak_regions.txt')
-    dat.mask_peak_regions(sample_dir,mod,q_powder,powder,qmask,show=True)
+        powder = hf['cake_integ'][()].sum(axis=0).sum(axis=0)
+        q_powder = hf['radial_units'][()].flatten()
+    path_peak_regions = os.path.join(sample_dir,'analysis','peak_regions.txt')
+    dat.mask_peak_regions(sample_dir,mod,q_powder,powder,path_peak_regions,show=True)
+    return path_peak_regions
 
-def show_some_baselines(baselines='simple'):
+def check_baselines( projection=0, datafile_pattern='.h5' ):
     """generates baselines for some random 1d images and plots them"""
-    return 0
+    bl_path_sample = os.path.join(sample_dir,'analysis','background_subtraction.py')
+    if not os.path.isfile( bl_path_sample ):
+        bl_path_module = hdl.get_file_path('textom',os.path.join('input','background_subtraction.py'))
+        os.makedirs(os.path.join(sample_dir,'analysis'), exist_ok=True )
+        shutil.copyfile(bl_path_module, bl_path_sample )
+    hdl.open_with_editor(bl_path_sample)
+    bl = msc.import_module_from_path('background_subtraction', bl_path_sample)
+    # load some data for testing
+    filelist = dat.get_data_list(sample_dir, datafile_pattern)
+    with h5py.File(os.path.join(sample_dir,'data_integrated',filelist[projection]), 'r') as hf:
+        q_1d = hf['radial_units'][()].flatten()
+        qmsk = np.logical_and(q_1d>=bl.q_min, q_1d<=bl.q_max )
+        data_1d = hf['cake_integ'][:,:,qmsk].sum(axis=1)
+        q_1d = q_1d[qmsk]
+    # get chosen peak regions for masking q-values
+    path_peak_regions = os.path.join(sample_dir,'analysis','peak_regions.txt')
+    if not os.path.isfile( path_peak_regions ):
+        mask_peak_regions()
+    peak_reg = dat.load_peak_regions(path_peak_regions)
+    q_mask, _ = dat.get_q_mask( q_1d, peak_reg )
+    # q_mask = np.ones_like(q_1d, dtype=bool)
+    # for (start, end) in peak_reg:
+    #     q_mask &= ~((q_1d >= start) & (q_1d <= end))
 
-def check_detector_mask():
+    happy = 'n'
+    while happy != 'y':
+        # make a graph with several random data and fits
+        f,ax = plt.subplots(3,3, figsize=(14,12), sharex=True,sharey=True)
+        k,s = 0,0
+        while k < 9:
+            pattern_1d = data_1d[np.random.randint(data_1d.shape[0])]
+            if dat.any_peak(pattern_1d[np.logical_and(q_1d > peak_reg[0,0], q_1d < peak_reg[0,1])],20) :
+                ax[k//3,k%3].plot(q_1d, pattern_1d, '-x')
+                ax[k//3,k%3].plot(q_1d, bln.no_baseline(q_1d, pattern_1d, q_mask), label='no baseline')
+                ax[k//3,k%3].plot(q_1d, bln.linear_baseline(q_1d, pattern_1d, q_mask), label='linear')
+                ax[k//3,k%3].plot(q_1d, bln.polynomial_baseline(q_1d, pattern_1d, q_mask, order=bl.order_polynom), label='polynomial')
+                ax[k//3,k%3].plot(q_1d, bln.chebyshev_baseline(q_1d,pattern_1d, q_mask, order=bl.order_chebyshev), label='chebyshev')
+                k += 1
+            s+=1 # safety counter
+            if s==5000:
+                break
+        # # collect all handles/labels
+        # handles, labels = ax[0,0].get_legend_handles_labels()
+        # by_label = dict(zip(labels, handles))
+        # f.legend(by_label.values(), by_label.keys())
+        ax[0,0].legend()
+        f.suptitle('Check baseline subtraction and choose the desired method')
+        f.tight_layout()
+        print('\tPlotted baselines with different modes. Choose the best one and enter parameters in your background_subtraction.py file. Type n to replot with new parameters')
+        happy = input('\t\thappy? (y/n) ')
+        if happy != 'y':
+            hdl.open_with_editor(bl_path_sample)
+            bl = msc.import_module_from_path('background_subtraction', bl_path_sample)
+
+def mask_detector_pixels( projection=0, datafile_pattern='.h5'):
     """shows the detector mask, maybe give a redo-option"""
-    return 0
+    detmask_path = os.path.join(sample_dir,'analysis','fit_detmask.txt')
+    # load some data for testing
+    filelist = dat.get_data_list(sample_dir, datafile_pattern)
+    with h5py.File(os.path.join(sample_dir,'data_integrated',filelist[projection]), 'r') as hf:
+        q = hf['radial_units'][()].flatten()
+        chi = hf['azimuthal_units'][()].flatten()
+        data = hf['cake_integ'][()].sum(axis=0)
+    # get chosen peak regions for masking q-values
+    path_peak_regions = os.path.join(sample_dir,'analysis','peak_regions.txt')
+    if not os.path.isfile( path_peak_regions ):
+        mask_peak_regions()
+    peak_reg = dat.load_peak_regions( path_peak_regions )
+    q_mask, q_mask_p = dat.get_q_mask( q, peak_reg )
+    powder_2D_masked = np.array([data[:,qmk].sum(axis=1) for qmk in q_mask_p])
+    
+    dat.mask_detector(sample_dir, detmask_path, powder_2D_masked, peak_reg, q, chi)
 
 def absorption_correction(absorption_tomogram=False, absorption_constant_voxel=False):
     abc.absorption_correction(sample_dir, mod,
         absorption_constant_voxel=absorption_constant_voxel, absorption_tomogram=absorption_tomogram)
 
-def preprocess_data( pattern='.h5', baselines='simple', use_ion=True ):
+def preprocess_data( proj_test=0, pattern='.h5', use_ion=True ):
     """Loads integrated data and pre-processes them for TexTOM
 
     Parameters
@@ -373,20 +466,24 @@ def preprocess_data( pattern='.h5', baselines='simple', use_ion=True ):
         'simple' to draw a straight line under each peak, (does not work if there is peak-overlap!)
         'polynomial' to fit the whole q-range minus the peaks with a 5th order polynomial (can
             handle peak-overlap, but data needs to be rather smooth where there is no peaks), 
-        'False' for no baselines (if your background is low), by default 'simple'
+        False for no baselines (if your background is low), by default 'simple'
     use_ion : bool, optional
         choose if you want to normalize data by the field 'ion' in the 
         data files, by default True
     """    
     if mod is None:
         make_model()
-    qmask = os.path.join(sample_dir,'analysis','peak_regions.txt')
-    detmask = os.path.join(sample_dir,'analysis','fit_detmask.txt')
-    dat.import_data( sample_dir, 
-                     pattern ,mod, 
-                     baselines=baselines,
-                     qmask_path=qmask,
-                     detmask_path=detmask,
+    path_peak_regions = os.path.join(sample_dir,'analysis','peak_regions.txt')
+    if not os.path.isfile( path_peak_regions ):
+        mask_peak_regions(proj_test, pattern)
+    path_background_subtraction = os.path.join(sample_dir,'analysis','background_subtraction.py')
+    if not os.path.isfile( path_background_subtraction ):
+        check_baselines(proj_test, pattern)
+    path_detector_mask = os.path.join(sample_dir,'analysis','fit_detmask.txt')
+    if not os.path.isfile( path_detector_mask ):
+        mask_detector_pixels(proj_test, pattern)
+    dat.import_data( sample_dir, pattern, mod, 
+                     path_background_subtraction, path_peak_regions, path_detector_mask,
                      geo_path=os.path.join(sample_dir,'analysis/geometry.py'),
                      use_ion=use_ion,
                      )
@@ -409,20 +506,7 @@ def make_fit( redo=True ):
     if (fit is None or redo):
         fit = fitting( sample_dir, mod )
 
-def check_orders(n_max=20, exclude_ghosts=True):
-    """Lists the sHSH orders available for the present symmetry
-
-    Parameters
-    ----------
-    n_max : int, optional
-        maximum order displayed, by default 20
-    """
-    if mod:
-        mod.get_orders(n_max,info=True, exclude_ghosts=exclude_ghosts)
-    else:
-        print('\tModel not defined.')
-
-def optimize( order=0, mode=0,
+def optimize( mode=4,
              proj='full', zero_peak=None,
              redo_fit=False,
              step_size_0=1,
@@ -461,10 +545,12 @@ def optimize( order=0, mode=0,
     """    
     global opt, results
     make_fit( redo=redo_fit )
-    fit.choose_projections( proj ) 
-    fit.set_orders( mod, order )
-    if mode==0:
+    fit.choose_projections( proj )
+    if mod.odf_mode =='hsh':
+        print( 'Optimizing 0 order intensities')
         fit.prepare_fit_c0(zero_peak)
+        # opimize c0 from average intensity
+        fit.C, opt = optimizer.optimize( fit, 0, tol=1e-3, minstep=1e-9, alg=alg, itermax=max_iter, step_size_0=step_size_0 ) 
     proj_c0 = False if mode==1 else True
     fit.C, opt = optimizer.optimize( fit, mode,
                 alg=alg, tol=tol, minstep=minstep, itermax=max_iter, project_c0=proj_c0, step_size_0=step_size_0 )
@@ -475,98 +561,91 @@ def optimize( order=0, mode=0,
         os.makedirs(os.path.join(sample_dir,'analysis','fits'), exist_ok=True)
         hdl.save_dict_to_h5( opt, os.path.join(sample_dir,'analysis','fits',opt['resname']))
 
-def optimize_auto( max_order=8, start_order=None, zero_peak=None,
-                tol_0 = 1e-7, tol_1 = 1e-3, tol_2 = 1e-4,
-                step_size_0=1,
-                minstep_0 = 1e-9, minstep_1 = 1e-9, minstep_2 = 1e-9,
-                max_iter = 500,
-                projections='full', alg='simple', adj_scal = False,
-                redo_fit=False,
-              ):
-    """Automated TexTOM reconstruction workflow
+# def optimize_auto( max_order=8, start_order=None, zero_peak=None,
+#                 tol_0 = 1e-7, tol_1 = 1e-3, tol_2 = 1e-4,
+#                 step_size_0=1,
+#                 minstep_0 = 1e-9, minstep_1 = 1e-9, minstep_2 = 1e-9,
+#                 max_iter = 500,
+#                 projections='full', alg='simple', adj_scal = False,
+#                 redo_fit=False,
+#               ):
+#     """Automated TexTOM reconstruction workflow
 
-    Parameters
-    ----------
-    max_order : int, optional
-        maximum HSH order to be used, by default 8    
-    start_order : int or None, optional
-        lowest order to be fitted, if None continues where you are standing, 
-        by default None
-    zero_peak : int or None
-        index of the peak you want to use for 0-order fitting (should be as
-        isotropic as possible), if None uses the whole dataset, default None
-    tol_0 : _type_, optional
-        tolerance for precision break criterion, mode 0, by default 1e-7
-    tol_1 : _type_, optional
-        tolerance for precision break criterion, mode 1, by default 1e-3
-    tol_2 : _type_, optional
-        tolerance for precision break criterion, mode 2, by default 1e-4
-    minstep_0 : _type_, optional
-        minimum stepsize in line search, mode 0, by default 1e-9
-    minstep_1 : _type_, optional
-        minimum stepsize in line search, mode 1, by default 1e-9
-    minstep_2 : _type_, optional
-        minimum stepsize in line search, mode 2, by default 1e-9
-    max_iter : int, optional
-        maximum number of iterations, by default 500
-    projections : str, optional
-        choose projections to be optimized: 'full', 'half', 'third', 'notilt', by default 'full'
-    alg : str, optional
-        choose algorithm between 'backtracking', 'simple', 'quadratic', 
-        by default 'simple'
-    adj_scal : bool, optional
-        rescales data after 0-order optimization, see adjust_data_scaling(), by default True
-    redo_fit : bool, optional
-        recalculate the fit object, by default False
-    """
-    global opt, results
-    make_fit( redo=redo_fit )
-    t0=time()
-    os.makedirs(os.path.join(sample_dir,'analysis','fits'), exist_ok=True)
-    fit.choose_projections( projections ) 
-    orders = mod.get_orders(n_max=max_order,info=False,exclude_ghosts=True)
-    if not isinstance(start_order, int):
-        if max(fit.ns)>0:
-            start_order = min(orders[orders>max(fit.ns)])
-        else:
-            start_order = 0
-    orders = orders[orders>=start_order]
-    if start_order==0:
-        print( 'Optimizing 0 order intensities')
-        fit.prepare_fit_c0(zero_peak)
-        # opimize c0 from average intensity
-        fit.C, opt = optimizer.optimize( fit, 0, tol=tol_0, minstep=minstep_0, alg=alg, itermax=max_iter, step_size_0=step_size_0 ) 
-        # plot.loss(opt['loss'],opt['resname'])
-        if adj_scal:
-            fit.adjust_data_scaling()
-            # fit.C, opt = optimizer.optimize( fit, 0, tol=tol_0, minstep=minstep_0, alg=alg ) 
-            # plot.loss(opt['loss'],opt['resname'])
-    print( 'Optimizing anisotropy')
-    for n in orders[orders>0]:
-        fit.set_orders( mod, n )
-        # optimize only highest
-        fit.C, opt = optimizer.optimize( fit, 3, tol=tol_1, minstep=minstep_1, alg=alg, project_c0=False, itermax=30, step_size_0=step_size_0 ) 
-        # plot.loss(opt['loss'],opt['resname'])
-        hdl.save_dict_to_h5( opt, os.path.join(sample_dir,'analysis','fits',opt['resname']))
+#     Parameters
+#     ----------
+#     max_order : int, optional
+#         maximum HSH order to be used, by default 8    
+#     start_order : int or None, optional
+#         lowest order to be fitted, if None continues where you are standing, 
+#         by default None
+#     zero_peak : int or None
+#         index of the peak you want to use for 0-order fitting (should be as
+#         isotropic as possible), if None uses the whole dataset, default None
+#     tol_0 : _type_, optional
+#         tolerance for precision break criterion, mode 0, by default 1e-7
+#     tol_1 : _type_, optional
+#         tolerance for precision break criterion, mode 1, by default 1e-3
+#     tol_2 : _type_, optional
+#         tolerance for precision break criterion, mode 2, by default 1e-4
+#     minstep_0 : _type_, optional
+#         minimum stepsize in line search, mode 0, by default 1e-9
+#     minstep_1 : _type_, optional
+#         minimum stepsize in line search, mode 1, by default 1e-9
+#     minstep_2 : _type_, optional
+#         minimum stepsize in line search, mode 2, by default 1e-9
+#     max_iter : int, optional
+#         maximum number of iterations, by default 500
+#     projections : str, optional
+#         choose projections to be optimized: 'full', 'half', 'third', 'notilt', by default 'full'
+#     alg : str, optional
+#         choose algorithm between 'backtracking', 'simple', 'quadratic', 
+#         by default 'simple'
+#     adj_scal : bool, optional
+#         rescales data after 0-order optimization, see adjust_data_scaling(), by default True
+#     redo_fit : bool, optional
+#         recalculate the fit object, by default False
+#     """
+#     global opt, results
+#     make_fit( redo=redo_fit )
+#     t0=time()
+#     os.makedirs(os.path.join(sample_dir,'analysis','fits'), exist_ok=True)
+#     fit.choose_projections( projections ) 
+#     orders = mod.get_orders(n_max=max_order,info=False,exclude_ghosts=True)
+#     if not isinstance(start_order, int):
+#         if max(fit.ns)>0:
+#             start_order = min(orders[orders>max(fit.ns)])
+#         else:
+#             start_order = 0
+#     orders = orders[orders>=start_order]
+#     if start_order==0:
+#         print( 'Optimizing 0 order intensities')
+#         fit.prepare_fit_c0(zero_peak)
+#         # opimize c0 from average intensity
+#         fit.C, opt = optimizer.optimize( fit, 0, tol=tol_0, minstep=minstep_0, alg=alg, itermax=max_iter, step_size_0=step_size_0 ) 
+#         # plot.loss(opt['loss'],opt['resname'])
+#         if adj_scal:
+#             fit.adjust_data_scaling()
+#             # fit.C, opt = optimizer.optimize( fit, 0, tol=tol_0, minstep=minstep_0, alg=alg ) 
+#             # plot.loss(opt['loss'],opt['resname'])
+#     print( 'Optimizing anisotropy')
+#     for n in orders[orders>0]:
+#         fit.set_orders( mod, n )
+#         # optimize only highest
+#         fit.C, opt = optimizer.optimize( fit, 3, tol=tol_1, minstep=minstep_1, alg=alg, project_c0=False, itermax=30, step_size_0=step_size_0 ) 
+#         # plot.loss(opt['loss'],opt['resname'])
+#         hdl.save_dict_to_h5( opt, os.path.join(sample_dir,'analysis','fits',opt['resname']))
 
-        # optimize all orders (mainly the 0th one though, the others converge slowly in this mode)
-        fit.C, opt = optimizer.optimize( fit, 2, tol=tol_2, minstep=minstep_2, alg=alg, itermax=30, step_size_0=step_size_0 ) 
-        # plot.loss(opt['loss'],opt['resname'])
+#         # optimize all orders (mainly the 0th one though, the others converge slowly in this mode)
+#         fit.C, opt = optimizer.optimize( fit, 2, tol=tol_2, minstep=minstep_2, alg=alg, itermax=30, step_size_0=step_size_0 ) 
+#         # plot.loss(opt['loss'],opt['resname'])
 
-    fit.C, opt = optimizer.optimize( fit, 3, tol=tol_1, minstep=minstep_1, alg=alg, project_c0=False, itermax=max_iter, step_size_0=step_size_0 ) 
-    hdl.save_dict_to_h5( opt, os.path.join(sample_dir,'analysis','fits',opt['resname']))
-    print(f'Total optimization time: {(time()-t0)/60} min')
-    results['optimization'] = opt['resname']
-    results['scaling'] = fit.C[:,0].reshape(mod.nVox)
-    check_projections_average()
-    check_projections_orientations()
-
-def adjust_data_scaling():
-    """Reestimates the data based on the assumption that normalization constants
-    contain noise. To be used after fitting the 0th order
-    """
-    global fit
-    fit.adjust_data_scaling()
+#     fit.C, opt = optimizer.optimize( fit, 3, tol=tol_1, minstep=minstep_1, alg=alg, project_c0=False, itermax=max_iter, step_size_0=step_size_0 ) 
+#     hdl.save_dict_to_h5( opt, os.path.join(sample_dir,'analysis','fits',opt['resname']))
+#     print(f'Total optimization time: {(time()-t0)/60} min')
+#     results['optimization'] = opt['resname']
+#     results['scaling'] = fit.C[:,0].reshape(mod.nVox)
+#     check_projections_average()
+#     check_projections_orientations()
 
 def list_opt(info=True):
     """Shows all stored optimizations
@@ -605,10 +684,10 @@ def load_opt( h5path = 'last', redo_fit=False, exclude_ghosts=True ):
     print(f'Loading {os.path.basename(h5path)}')
     global fit,mod,opt, results
     make_fit( redo=redo_fit )
-    i0 = h5path.find('nmax')
-    i1 = h5path.find('_', i0)
-    nmax = int( h5path[i0+4 : i1] )
-    fit.set_orders(mod,nmax, exclude_ghosts=exclude_ghosts)
+    # i0 = h5path.find('nmax')
+    # i1 = h5path.find('_', i0)
+    # nmax = int( h5path[i0+4 : i1] )
+    # fit.set_orders(mod,nmax, exclude_ghosts=exclude_ghosts)
     opt = hdl.load_h5_to_dict( h5path )
     fit.C = opt['c'].astype(data_type)
     results['optimization'] = opt['resname']
@@ -639,8 +718,9 @@ def check_fit_average( ):
     Parameters
     ----------
     """    
+    ## this plot should be per tilt angle with the rotation angle at the bottom
     if not hasattr(fit, 'data_0D'):
-        fit.prepare_fit_c0
+        fit.prepare_fit_c0()
     plt.figure()
     plt.plot(fit.data_0D, label='data')
     plt.plot(fit.projections_c0(fit.C),label='fit')
@@ -741,14 +821,14 @@ def check_projections_orientations( G=None ):
     else:
         fov = mod.fov
     for g in np.atleast_1d( G ):
-        title = f'Projection {g}: kap {np.round(mod.Kappa[g]*180/np.pi,1)}, ome {np.round(mod.Omega[g]*180/np.pi,1)}'
+        title = f'Projection {g}: kap {np.round(mod.Kappa[g],1)}, ome {np.round(mod.Omega[g],1)}'
         ori_dat, ori_fit = fit.check_orientations(g)
         plot.compare_projection_ori( 
             ori_dat.reshape((ori_dat.shape[0],*fov)), 
             ori_fit.reshape((ori_dat.shape[0],*fov)),
             fit.q, title=title, )
         
-def calculate_orientation_statistics( truncate_expansion=False ):
+def calculate_orientation_statistics( ):
     """Calculates prefered orientations and stds and saves them to results dict
 
     """
@@ -758,17 +838,18 @@ def calculate_orientation_statistics( truncate_expansion=False ):
         return
     results['scaling'] = fit.C[:,0].reshape(mod.nVox)
     print('\tExtracting mean orientations')
-    g_pref = mod.g_ML_sample( fit.C[mod.mask_voxels], truncate=truncate_expansion )
+    g_pref = mod.preferred_orientations( fit.C[mod.mask_voxels] )
     results['g_pref'] = mod.insert_sparse_tomogram(g_pref)
     print('\tExtracting stds')
     stds = mod.std_sample( fit.C[mod.mask_voxels], g_pref )
     results['std']=mod.insert_sparse_tomogram(stds)
     # export vtk-file for paraview
-    a_pref, b_pref, c_pref = mod.abc_pref_sample( g_pref )
+    crystal_data = cry.parse_cif(mod.cr.cifPath)
+    a_pref, b_pref, c_pref = mod.abc_pref_sample( g_pref, crystal_data['lattice_vectors'] )
     results['a_pref'] = mod.insert_sparse_tomogram(a_pref)
     results['b_pref'] = mod.insert_sparse_tomogram(b_pref)
     results['c_pref'] = mod.insert_sparse_tomogram(c_pref)
-    results['inv_pf'] = orx.inverse_pf_stack(results['g_pref'], mod)
+    results['inv_pf'] = orx.inverse_pf_stack(results['g_pref'], mod.symmetry, mod.mask_voxels)
     print("\tAdded to results: 'std', 'g_pref', 'a_pref', 'b_pref', 'c_pref', 'inv_pf'")
 
 def calculate_order_parameters(axis=(0,0,1)):
@@ -777,9 +858,9 @@ def calculate_order_parameters(axis=(0,0,1)):
         print('model/fit not defined!')
         return
     results['scaling'] = fit.C[:,0].reshape(mod.nVox)
-    directions = mod.directions_from_orientations( mod.Gc, np.array(axis, data_type) )
-    order_parameters, directors = ord.order_parameters_parallel( 
-        directions, fit.C[mod.mask_voxels], mod.sHSHs, mod.V_fz )
+    directions, Gc = mod.directions_from_orientations( np.array(axis, data_type) )
+    odfs = mod.odf_module.get_odf_batch(fit.C[mod.mask_voxels], Gc, mod.symmetry, mod.odf_parameter)
+    order_parameters, directors = ord.order_parameters_parallel( directions, odfs )
     results['order_parameters'] = mod.insert_sparse_tomogram(order_parameters)
     results['directors'] = mod.insert_sparse_tomogram(directors)
     print("\tAdded to results: 'order_parameters', 'directors'")
@@ -797,9 +878,12 @@ def calculate_segments(thresh=10, min_segment_size=30, max_segments_number=31):
         maximum number of segments (ordered by size), by default 32
     """
     global results
-    with h5py.File(os.path.join(sample_dir,'analysis','diffractlets.h5'), 'r') as hf:
-        ppgrp = hf['symmetry'][()].decode('utf-8')
-    mx,my,mz,mori = seg.mori_3d( results['g_pref'], ppgrp )
+    cr = msc.import_module_from_path('crystal', os.path.join(sample_dir,'analysis','crystal.py'))
+    difflets_path = os.path.join(sample_dir,'analysis',f'difflets_{cr.odf_mode}.h5')
+    if os.path.isfile(difflets_path):
+        with h5py.File(difflets_path, 'r') as hf:
+            symmetry = hf['symmetry'][()].decode('utf-8')
+    mx,my,mz,mori = seg.mori_3d( results['g_pref'], symmetry )
     plt.figure()
     plt.hist(mori[mori<3.14]*180/np.pi,50)
     plt.plot([thresh,thresh],[mori.shape[0],mori.shape[0]],'--')
@@ -873,10 +957,13 @@ def show_slice_ipf( h, plane='z' ):
     plane : str, optional
         slice direction: x/y/z, by default 'z'
     """
-    with h5py.File(os.path.join(sample_dir,'analysis/diffractlets.h5'), 'r') as hf:
-        ppgrp = hf['symmetry'][()].decode('utf-8')
+    cr = msc.import_module_from_path('crystal', os.path.join(sample_dir,'analysis','crystal.py'))
+    difflets_path = os.path.join(sample_dir,'analysis',f'difflets_{cr.odf_mode}.h5')
+    if os.path.isfile(difflets_path):
+        with h5py.File(difflets_path, 'r') as hf:
+            symmetry = hf['symmetry'][()].decode('utf-8')
     ipf_slice,_ = _slice('g_pref', h, plane)
-    orx.inverse_pf_map( ipf_slice, ppgrp )
+    orx.inverse_pf_map( ipf_slice, symmetry )
 
 def show_slice_directions( h, plane='z', direction='c' ):
     """Plots an inverse pole figure of a sample slice
@@ -906,11 +993,14 @@ def show_volume_ipf( plane='z', save=False, show=True ):
     show : bool, optional
         open the figure upon calling the function, by default True
     """
-    with h5py.File(os.path.join(sample_dir,'analysis/diffractlets.h5'), 'r') as hf:
-        ppgrp = hf['symmetry'][()].decode('utf-8')
+    cr = msc.import_module_from_path('crystal', os.path.join(sample_dir,'analysis','crystal.py'))
+    difflets_path = os.path.join(sample_dir,'analysis',f'difflets_{cr.odf_mode}.h5')
+    if os.path.isfile(difflets_path):
+        with h5py.File(difflets_path, 'r') as hf:
+            symmetry = hf['symmetry'][()].decode('utf-8')
     if save:
         save = os.path.join(sample_dir, 'results', 'images', f'ipf_{plane}_{msc.timestring()}.gif')
-    orx.plot_ipf_movie(results['inv_pf'], ppgrp, plane, save_vid_path=save, show=show)
+    orx.plot_ipf_movie(results['inv_pf'], symmetry, plane, save_vid_path=save, show=show)
 
 def show_histogram(x, nbins=50, cut=0.1, segments=None, save=False):
     """plots a histogram of a result parameter
@@ -1015,28 +1105,35 @@ def show_correlations( x, y, nbins=50, cut=(0.1,0.1), segments=None, save=False
                 save='png'
                 f.savefig(os.path.join(sample_dir,'results','images',f'correlations_{x}_{y}_{msc.timestring()}.{save}'))
 
-def show_voxel_odf( x,y,z, num_samples=1000, kernel=None, info=True ):
-    """Show a 3D plot of the ODF in the chosen voxel
+def show_voxel_odf( x,y,z, num_samples=1000, representation='quaternion', info=True ):
+    # """Show a 3D plot of the ODF in the chosen voxel
 
-    Parameters
-    ----------
-    x : int
-        voxel x-coordinate
-    y : int
-        voxel y-coordinate
-    z : int
-        voxel z-coordinate
-    num_samples : int/float, optional
-        number of samples for plot generation, by default 1000
-    kernel : float / 'auto' / None, optional
-        choose the strength k of the kernel or 'auto' to refine k,
-        if None clips the distribution instead, by default None
-    """
+    # Parameters
+    # ----------
+    # x : int
+    #     voxel x-coordinate
+    # y : int
+    #     voxel y-coordinate
+    # z : int
+    #     voxel z-coordinate
+    # representation : 'quaternion'|'euler'|'rodrigues'
+    #   different styles
+    # """
     iv = np.ravel_multi_index((x,y,z), mod.nVox)
     if np.any(fit.C[iv]):
+        Gc, odf = mod.odfFromC(fit.C[iv])
         title = f'Voxel {x}/{y}/{z}'
-        orx.odf_cloud(mod, fit.C[iv], num_samples=num_samples, title=title, 
-                      kernel=kernel, odf_info=info )
+        if representation=='euler':
+            euler_angles = rot.EulerfromQ(rot.QfromOTP(Gc))
+            plot.plot_odf_slices_euler(euler_angles, odf)
+        if representation=='rodrigues':
+            rodrigues_par = rot.RodfromOTP(Gc)
+            plot.plot_odf_slices_rodrigues(rodrigues_par,odf)
+        else: #'quaternion'
+            orx.plot_odf( rot.QfromOTP(Gc), 
+                        odf, mod.symmetry, num_samples=num_samples, title=title )
+        # orx.odf_cloud(mod, fit.C[iv], num_samples=num_samples, title=title, 
+        #               kernel=kernel, odf_info=info )
     else:
         print('\tNo ODF in this voxel')
 
@@ -1064,8 +1161,11 @@ def show_voxel_polefigure( x,y,z, hkl=(1,0,0), mode='density',
     iv = np.ravel_multi_index((x,y,z), mod.nVox)
     title = f'Voxel {x}/{y}/{z}, hkl ({hkl[0]},{hkl[1]},{hkl[2]})'
     if np.any(fit.C[iv]):
-        orx.plot_pole_figure(mod, fit.C[iv], hkl, mode=mode, title=title,
-                             num_samples=num_samples, alpha=alpha, recenter=recenter )
+        # calculate odf, recenter if necessary, then feed the odf to the function
+        Gc, odf = mod.odfFromC(fit.C[iv], recenter=recenter)
+
+        orx.plot_pole_figure(Gc, odf, mod.symmetry, hkl, mode=mode, title=title,
+                             num_samples=num_samples, alpha=alpha )
     else:
         print('\tNo ODF in this voxel')
 
